#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <regex.h>
#include <string.h>

#include "udm_config.h"
#include "udm_common.h"
#include "udm_utils.h"
#include "udm_filter.h"

/************ Allow/Disallow/CheckOnly  stuff *****************/

typedef struct filter_struct {
	regex_t filter;
	int     filter_type;
	int	reverse;
	char	*regstr;
} UDM_FILTER;


int UdmAddFilter(UDM_ENV * Conf,char * filter,int filter_type,int reverse){
#define ERRSTRSIZE 100
char regerrstr[UDMSTRSIZ]="";
int err;
UDM_FILTER * Filter;
	if(Conf->nfilters>=Conf->mfilters){
		if(Conf->mfilters){
			Conf->mfilters+=16;
			Conf->Filter=(UDM_FILTER *)realloc(Conf->Filter,Conf->mfilters*sizeof(UDM_FILTER));
		}else{
			Conf->mfilters=16;
			Conf->Filter=(UDM_FILTER *)malloc(Conf->mfilters*sizeof(UDM_FILTER));
		}
	}
	Filter=(UDM_FILTER*)Conf->Filter;
	err=regcomp(&(Filter[Conf->nfilters].filter),filter,REG_EXTENDED|REG_ICASE);
	if(err){
		regerror(err, &(Filter[Conf->nfilters].filter), regerrstr, ERRSTRSIZE);
		Conf->errcode=1;
		sprintf(Conf->errstr,"Wrong regex in config file: %s: %s", filter,regerrstr);
		regfree(&(Filter[Conf->nfilters].filter));
		return(1);
	}
	Filter[Conf->nfilters].filter_type=filter_type;
	Filter[Conf->nfilters].reverse=reverse;
	Filter[Conf->nfilters].regstr=strdup(filter);
	Conf->nfilters++;
	return(0);
}
void UdmFreeFilters(UDM_ENV * Conf){
int i;
UDM_FILTER * Filter;
	Filter=(UDM_FILTER*)Conf->Filter;
	for(i=0; i < Conf->nfilters; i++){
		regfree(&(Filter[i].filter));
		free(Filter[i].regstr);
	}
	UDM_FREE(Conf->Filter);
	Conf->nfilters=0;
	Conf->mfilters=0;
}
int UdmFindFilter(UDM_ENV * Conf,char * param,char * reason){
int i;
#define NS 10
regmatch_t subs[NS];
int err;
UDM_FILTER * Filter;

	Filter=(UDM_FILTER *)Conf->Filter;
	reason[0]=0;
	for(i=0;i<Conf->nfilters;i++){
		err=regexec(&(Filter[i].filter),param,NS,subs,0);
		if((Filter[i].reverse)&&(err))
			break;
		if((!Filter[i].reverse)&&(!err))
			break;
	}
	if(i<Conf->nfilters){
		switch(Filter[i].filter_type){
			case UDM_ALLOW:		strcpy(reason,"Allow");break;
			case UDM_DISALLOW:	strcpy(reason,"Disallow");break;
			case UDM_HEAD:		strcpy(reason,"CheckOnly");break;
			case UDM_HREFONLY:	strcpy(reason,"HrefOnly");break;
			default:		strcpy(reason,"Unknown");break;
		}
		strcpy(UDM_STREND(reason),Filter[i].reverse?"NoMatch":"");
		strcpy(UDM_STREND(reason),Filter[i].regstr?Filter[i].regstr:"");
		return(Filter[i].filter_type);
	}
	strcpy(reason,"Allow by default");
	return(UDM_ALLOW);
}
