#include <stdio.h>
#include <stdlib.h>

#include "udm_config.h"
#include "udm_common.h"
#include "udm_utils.h"
#include "udm_charset.h"
#include "udm_indexer.h"
#include "udm_proto.h"
#include "udm_parseurl.h"
#include "udm_hrefs.h"
#include "udm_server.h"

#define DEFAULT_PROXY_PORT	3128

int UdmFollowType(char * follow){
	if(!strcasecmp(follow,"no"))return(UDM_FOLLOW_PAGE);
	if(!strcasecmp(follow,"page"))return(UDM_FOLLOW_PAGE);
	if(!strcasecmp(follow,"yes"))return(UDM_FOLLOW_PATH);
	if(!strcasecmp(follow,"path"))return(UDM_FOLLOW_PATH);
	if(!strcasecmp(follow,"site"))return(UDM_FOLLOW_SITE);
	if(!strcasecmp(follow,"world"))return(UDM_FOLLOW_WORLD);
	return(UDM_FOLLOW_UNKNOWN);
}

/* return values: 0 on success, non-zero on error */
int UdmAddServer(UDM_ENV * Conf,UDM_SERVER * srv, int load_flags){
	int res;
	UDM_URL from;

	/* Check whether valid URL is passed */
	if((srv->url[0])&&(res=UdmParseURL(&from,srv->url))){
		return(res);
	}
	if((srv->url[0])&&(load_flags&UDM_FLAG_ADD_SERV)){
		UdmAddHref(Conf,srv->url,0,0,0);
		switch(srv->follow){
			char * s;
			case UDM_FOLLOW_PATH:
				/* Cut after last '/' */
				if((s=strrchr(srv->url,'/')))
					*(s+1)='\0';
				break;

			case UDM_FOLLOW_SITE:
				if(from.hostinfo[0]){
					/* Cut after hostinfo */
					int len;
					len=strlen(srv->url);
					UDM_FREE(srv->url);
					srv->url=(char*)malloc(len);
					sprintf(srv->url,"%s://%s/",from.schema,from.hostinfo);

				}else{
					/* Cut after first '/' */
					if((s=strchr(srv->url,'/')))
						*(s+1)='\0';
				}
				break;
				
			case UDM_FOLLOW_PAGE: 
			case UDM_FOLLOW_WORLD:
			default:
				break;
		}

		/* Add robots.txt for HTTP*/
		if((!strcmp(from.schema,"http"))&&(srv->use_robots)){
			char str[UDMSTRSIZ];
			sprintf(str,"%s://%s/%s",from.schema,from.hostinfo,"robots.txt");
			UdmAddHref(Conf,str,0,0,0);
		}
	}
	if (!strcmp(from.schema, "news")) {
		char *c, *cc;
		/* Cat server name to remove group names */
		/* This is because group names do not    */
		/* present in message URL                */
		c=srv->url+7;
		cc=strchr(c,'/');
		if(cc)*(cc+1)='\0';
	}
	if(Conf->nservers>=Conf->mservers){
		if(Conf->mservers){
			Conf->mservers+=16;
			Conf->Server=(UDM_SERVER *)realloc(Conf->Server,Conf->mservers*sizeof(UDM_SERVER));
		}else{
			Conf->mservers=16;
			Conf->Server=(UDM_SERVER *)malloc(Conf->mservers*sizeof(UDM_SERVER));
		}
	}
	Conf->Server[Conf->nservers].url=strdup(srv->url);
	Conf->Server[Conf->nservers].charset=srv->charset?strdup(srv->charset):NULL;
	Conf->Server[Conf->nservers].basic_auth=srv->basic_auth?strdup(srv->basic_auth):NULL;
	Conf->Server[Conf->nservers].proxy_basic_auth=srv->proxy_basic_auth?strdup(srv->proxy_basic_auth):NULL;
	Conf->Server[Conf->nservers].htdb_list=srv->htdb_list?strdup(srv->htdb_list):NULL;
	Conf->Server[Conf->nservers].htdb_doc=srv->htdb_doc?strdup(srv->htdb_doc):NULL;
	Conf->Server[Conf->nservers].user=srv->user?strdup(srv->user):NULL;
	Conf->Server[Conf->nservers].passwd=srv->passwd?strdup(srv->passwd):NULL;
	Conf->Server[Conf->nservers].category=srv->category?strdup(srv->category):NULL;
	Conf->Server[Conf->nservers].tag=srv->tag?strdup(srv->tag):NULL;
	Conf->Server[Conf->nservers].mirror_root=srv->mirror_root?strdup(srv->mirror_root):NULL;
	Conf->Server[Conf->nservers].mirror_headers=srv->mirror_headers?strdup(srv->mirror_headers):NULL;
	
	strncpy(Conf->Server[Conf->nservers].lang,srv->lang,2);
	
	if(srv->proxy){
		char *s;
		Conf->Server[Conf->nservers].proxy=strdup(srv->proxy);
		if((s=strchr(Conf->Server[Conf->nservers].proxy,':'))){
			*s=0;
			Conf->Server[Conf->nservers].proxy_port=atoi(s+1)?atoi(s+1):DEFAULT_PROXY_PORT;
		}else{
			Conf->Server[Conf->nservers].proxy_port=DEFAULT_PROXY_PORT;
		}
	}else{
		Conf->Server[Conf->nservers].proxy=NULL;
		Conf->Server[Conf->nservers].proxy_port=0;
	}
	Conf->Server[Conf->nservers].period=srv->period;
	Conf->Server[Conf->nservers].net_errors=0;
	Conf->Server[Conf->nservers].maxhops=srv->maxhops;
	Conf->Server[Conf->nservers].index=srv->index;
	Conf->Server[Conf->nservers].follow=srv->follow;
	Conf->Server[Conf->nservers].deletebad=srv->deletebad;
	Conf->Server[Conf->nservers].use_robots=srv->use_robots;
	Conf->Server[Conf->nservers].bodyweight=srv->bodyweight;
	Conf->Server[Conf->nservers].titleweight=srv->titleweight;
	Conf->Server[Conf->nservers].urlweight=srv->urlweight;
	Conf->Server[Conf->nservers].urlhostweight=srv->urlhostweight;
	Conf->Server[Conf->nservers].urlpathweight=srv->urlpathweight;
	Conf->Server[Conf->nservers].urlfileweight=srv->urlfileweight;
	Conf->Server[Conf->nservers].descweight=srv->descweight;
	Conf->Server[Conf->nservers].keywordweight=srv->keywordweight;
	Conf->Server[Conf->nservers].max_net_errors=srv->max_net_errors;
	Conf->Server[Conf->nservers].read_timeout=srv->read_timeout;
	Conf->Server[Conf->nservers].delete_no_server=srv->delete_no_server;
	Conf->Server[Conf->nservers].correct_factor=srv->correct_factor;
	Conf->Server[Conf->nservers].incorrect_factor=srv->incorrect_factor;
	Conf->Server[Conf->nservers].number_factor=srv->number_factor;
	Conf->Server[Conf->nservers].alnum_factor=srv->alnum_factor;
	Conf->Server[Conf->nservers].min_word_length=srv->min_word_length;
	Conf->Server[Conf->nservers].max_word_length=srv->max_word_length;
	Conf->Server[Conf->nservers].use_mirror=srv->use_mirror;
	Conf->Server[Conf->nservers].use_clones=srv->use_clones;
	Conf->Server[Conf->nservers].check_mp3_tag=srv->check_mp3_tag;
	Conf->Server[Conf->nservers].check_only_mp3_tag=srv->check_only_mp3_tag;
	Conf->Server[Conf->nservers].net_error_delay_time=srv->net_error_delay_time;
	Conf->nservers++;
	return(0);
}

void UdmFreeServers(UDM_ENV * Conf){
int i;
	for(i=0;i<Conf->nservers;i++){
		UDM_FREE(Conf->Server[i].url);
		UDM_FREE(Conf->Server[i].charset);
		UDM_FREE(Conf->Server[i].basic_auth);
		UDM_FREE(Conf->Server[i].proxy_basic_auth);
		UDM_FREE(Conf->Server[i].htdb_list);
		UDM_FREE(Conf->Server[i].htdb_doc);
		UDM_FREE(Conf->Server[i].proxy);
		UDM_FREE(Conf->Server[i].category);
		UDM_FREE(Conf->Server[i].tag);
		UDM_FREE(Conf->Server[i].mirror_root);
		UDM_FREE(Conf->Server[i].mirror_headers);
	}
	Conf->nservers=Conf->mservers=0;
	UDM_FREE(Conf->Server);
}

UDM_SERVER * UdmFindServer(UDM_ENV * Conf,char *url){
int i;
char *r;

	if(!Conf->Server)return(NULL);
	url=UdmRemove2Dot(url);
	r=strstr(url,"/robots.txt");
	/* If it's a robot.txt, cut to hostinfo and find result */
	if((r)&&(!strcmp(r,"/robots.txt"))){
		r++;*r=0;
		for(i=0;i<Conf->nservers;i++){
			if(!UDM_STRNCMP(url,Conf->Server[i].url)){
				*r='r';
				return(&Conf->Server[i]);
			}
		}
		*r='r';
	}else{
		for(i=0;i<Conf->nservers;i++){
			if(!UDM_STRNCMP(url,Conf->Server[i].url))
				return(&Conf->Server[i]);
		}
	}
	return(0);
}

static int cmpserver(const void *s1,const void *s2){
	return(strlen(((UDM_SERVER*)s2)->url)-strlen(((UDM_SERVER*)s1)->url));
}
void UdmSortServers(UDM_ENV * Conf){
	/*  Long name should be found first    */
	/*  to allow different options         */
	/*  for server and it's subdirectories */
	qsort((void*)Conf->Server,Conf->nservers,sizeof(UDM_SERVER),cmpserver);
}
int UdmInitServer(UDM_SERVER * srv){
	srv->url=NULL;
	srv->proxy=NULL;
	srv->basic_auth=NULL;
	srv->proxy_basic_auth=NULL;
	srv->charset=UDM_CHARSET_USASCII;
	srv->htdb_list=NULL;
	srv->htdb_doc=NULL;
	srv->category=NULL;
	srv->proxy_port=0;
	srv->period=UDM_DEFAULT_REINDEX_TIME;
	srv->tag=NULL;
	srv->mirror_root=NULL;
	srv->mirror_headers=NULL;
	srv->net_errors=0;
	srv->max_net_errors=UDM_MAXNETERRORS;
	srv->read_timeout=UDM_READ_TIMEOUT;
	srv->delete_no_server=1;
	srv->maxhops=UDM_DEFAULT_MAX_HOPS;
	srv->index=1;
	srv->follow=UDM_FOLLOW_PATH;
	srv->deletebad=0;
	srv->use_robots=1;
	srv->bodyweight=1;
	srv->titleweight=2;
	srv->keywordweight=2;
	srv->descweight=2;
	srv->urlweight=0;
	srv->urlhostweight=0;
	srv->urlpathweight=0;
	srv->urlfileweight=0;
	srv->correct_factor=1;
	srv->incorrect_factor=1;
	srv->number_factor=1;
	srv->alnum_factor=1;
	srv->min_word_length=1;
	srv->max_word_length=32;
	srv->use_mirror=UDM_MIRROR_NO;
	srv->use_clones=1;
	srv->net_error_delay_time=UDM_DEFAULT_NET_ERROR_DELAY_TIME;
	srv->lang[0]=0;
	srv->user=NULL;
	srv->passwd=NULL;
	return(0);
}
